#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <cuda.h>
#include <sys/time.h>
#include <math.h>

#define CHK_CUDA(e) {if (e != cudaSuccess) {fprintf(stderr,"Error: %s\n", cudaGetErrorString(e)); exit(-1);}}

__global__ void cuda_reverse(const float *xd, 
                                   float *yd, 
                             const int n) {

	int i = threadIdx.x + blockIdx.x*blockDim.x;
	if (i<n) {
		yd[n-(i+1)] = xd[i];
	}
	return;
}

__global__ void cuda_reverse_coalesced(const float *xd, 
                                             float *yd, 
                                       const int n) {

    extern __shared__ float blockdata[];
	int iin = threadIdx.x + blockIdx.x*blockDim.x;
    int outblock = gridDim.x - (blockIdx.x + 1);
    int iout = threadIdx.x + outblock*blockDim.x;
	if (iin<n) {
		blockdata[threadIdx.x] = xd[iin];
        __syncthreads();
        yd[iout] = blockdata[blockDim.x - (threadIdx.x+1)];
	}
	return;
}

void cpu_reverse(const float *x, float *y, int n) {

	int i;
	for (i=0;i<n;i++) {
		y[n-(i+1)] = x[i];
	}
	return;
}

int get_options(int argc, char **argv, int *n, int *nb, int *niters);
void tick(struct timeval *timer);
double tock(struct timeval *timer);

int main(int argc, char **argv) {
	int n=1000;
	int nblocks=10;
	int err;
	float *x, *y, *ycuda, *ycuda2;
	float *xd, *yd;
	double abserr, abserr2;
	int blocksize;
    int i;
    int niters=1;
    struct timeval cputimer;
    struct timeval gputimer;
    struct timeval gputimer2;
    double cputime, gputime, gputime2;

	err = get_options(argc, argv, &n, &nblocks, &niters);
	if (err) return 0;

	x = (float *)malloc(n*sizeof(float));
	if (!x) return 1;
		
	y = (float *)malloc(n*sizeof(float));
	if (!y) {free(x); return 1;}

	ycuda = (float *)malloc(n*sizeof(float));
	if (!ycuda) {free(y); free(x); return 1;}

	ycuda2 = (float *)malloc(n*sizeof(float));
	if (!ycuda2) {free(ycuda2); free(y); free(x); return 1;}


    for (i=0; i<n; i++) {
        x[i] = 1.*i*i;
        ycuda[i] = 0.;
        ycuda2[i]= 0.;
    }

	/* run CPU code */

    tick(&cputimer);
    for (i=0; i<niters;i++)
        cpu_reverse(x, y, n);
    cputime = tock(&cputimer);

	/* run unstrided GPU code */
	CHK_CUDA( cudaMalloc(&xd, n*sizeof(float)) );
	CHK_CUDA( cudaMalloc(&yd, n*sizeof(float)) );
	CHK_CUDA( cudaMemcpy(yd, ycuda, n*sizeof(float), cudaMemcpyHostToDevice) );

    tick(&gputimer);
	CHK_CUDA( cudaMemcpy(xd, x, n*sizeof(float), cudaMemcpyHostToDevice) );

	blocksize = (n+nblocks-1)/nblocks;
    for (i=0; i<niters; i++) {
        cuda_reverse<<<nblocks, blocksize>>>(xd, yd, n);
    }
	CHK_CUDA( cudaMemcpy(ycuda, yd, n*sizeof(float), cudaMemcpyDeviceToHost) );
    gputime = tock(&gputimer);
	CHK_CUDA( cudaFree(xd) );
	CHK_CUDA( cudaFree(yd) );
	
	/* run strided GPU code */

	CHK_CUDA( cudaMalloc(&xd, n*sizeof(float)) );
	CHK_CUDA( cudaMalloc(&yd, n*sizeof(float)) );
	CHK_CUDA( cudaMemcpy(yd, ycuda2, n*sizeof(float), cudaMemcpyHostToDevice) );

    tick(&gputimer2);
	CHK_CUDA( cudaMemcpy(xd, x, n*sizeof(float), cudaMemcpyHostToDevice) );

	blocksize = (n+nblocks-1)/nblocks;
    for (i=0; i<niters; i++) {
        cuda_reverse_coalesced<<<nblocks, blocksize, blocksize*sizeof(float)>>>(xd, yd, n);
    }
	CHK_CUDA( cudaMemcpy(ycuda2, yd, n*sizeof(float), cudaMemcpyDeviceToHost) );
    gputime2 = tock(&gputimer2);
	
	CHK_CUDA( cudaFree(xd) );
	CHK_CUDA( cudaFree(yd) );

    abserr  = 0.;
    abserr2 = 0.;
    for (i=0;i<n;i++) {
        abserr  += fabs(ycuda[i] - y[i]);
        abserr2 += fabs(ycuda2[i] - y[i]);
    }

    printf("For run with n = %d, nblocks = %d, blocksize = %d, iters=%d,\n", n, nblocks, blocksize, niters);
    printf("CPU time  = %lg millisec.\n", cputime*1000.);
    printf("GPU time  = %lg millisec, diff = %lf.\n", gputime*1000., abserr);
    printf("GPU2 time = %lg millisec, diff = %lf.\n", gputime2*1000., abserr2);

    if (n < 20) {
        for (i=0; i<n; i++) printf("%6.1f ",y[i]);
        puts("");
        for (i=0; i<n; i++) printf("%6.1f ",ycuda[i]);
        puts("");
        for (i=0; i<n; i++) printf("%6.1f ",ycuda2[i]);
        puts("");
    }

    free(x);
    free(y);
    free(ycuda);
    free(ycuda2);
	return 0;
}


int get_options(int argc, char **argv, int *n, int *nb, int *niters) {

  const struct option long_options[] = {
    {"nvals"     , required_argument, 0, 'n'},
    {"nblocks"   , required_argument, 0, 'B'},
    {"niters", required_argument, 0, 'I'},
    {"help",      no_argument, 0, 'h'},
    {0, 0, 0, 0}};

  char c;
  int option_index;
  int tempint;

  while (1) {
    c = getopt_long(argc, argv, "n:B:a:b:s:h", long_options, &option_index);
    if (c == -1) break;
 
    switch(c) {
      case 'n': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 500000) {
            fprintf(stderr,"%s: Cannot use number of points %s;\n  Using %d\n", argv[0], optarg, *n);
          } else {
            *n = tempint;
          }
          break;

      case 'B': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 1000 || tempint > *n) {
            fprintf(stderr,"%s: Cannot use number of blocks %s;\n  Using %d\n", argv[0], optarg, *nb);
          } else {
            *nb = tempint;
          }
          break;

      case 'I': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 1000 || tempint > *n) {
            fprintf(stderr,"%s: Cannot use number of iterations %s;\n  Using %d\n", argv[0], optarg, *niters);
          } else {
            *niters = tempint;
          }
          break;

      case 'h':
          puts("Reverses an array on the GPU.");
          puts("Options: ");
          puts("    --nvals=N      (-n N): Set the number of values in y,x.");
          puts("    --nblocks=N    (-B N): Set the number of blocks used.");
          puts("    --niters=N     (-I X): Set number of iterations to calculate.");
          puts("");
          return +1;
        }
    }

    return 0;
}

void tick(struct timeval *timer) {
    gettimeofday(timer, NULL);
}

double tock(struct timeval *timer) {
    struct timeval now;
    gettimeofday(&now, NULL);
    return (now.tv_usec-timer->tv_usec)/1.0e6 + (now.tv_sec - timer->tv_sec);
}
