#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdint.h>
#include <cuda.h>
#include <sys/time.h>


#define CHK_CUDA(e) {if (e != cudaSuccess) {fprintf(stderr,"Error: %s\n", cudaGetErrorString(e)); exit(-1);} }
#define CHK_ERROR() {cudaError e=cudaGetLastError(); CHK_CUDA(e); }


/* transliteration from the CPU code */
__global__ void cuda_smoothimage(uint8_t *data,
                                 uint8_t *smootheddata,
                                 const int nhalos, const int rows, const int cols) {

    const double sig = 2./9.*(nhalos*nhalos);
    const double pi = 4.*atan(1.);
    const int rowsize=cols+2*nhalos;

    int r = threadIdx.y + blockIdx.y*blockDim.y;
    int c = threadIdx.x + blockIdx.x*blockDim.x;

    if (r<rows && c<cols) {
        r += nhalos;
        c += nhalos;
        double mean = 0.;
        for (int i=r-nhalos; i<=r+nhalos; i++) {
             for (int j=c-nhalos; j<=c+nhalos; j++) {
                 double d  = 1.0*data[i*rowsize+j];
                 double e = exp(-(double)((i-r)*(i-r)+(j-c)*(j-c))/sig);
                 mean += d*e;
             }
        }
        smootheddata[r*rowsize + c] = (uint8_t)round(mean/(sig*pi));
    }
    return;
}

__global__ void cuda_smoothimage_uncoallesced(uint8_t *data,
                                 uint8_t *smootheddata,
                                 const int nhalos, const int rows, const int cols) {

    const double sig = 2./9.*(nhalos*nhalos);
    const double pi = 4.*atan(1.);
    const int rowsize=cols+2*nhalos;

    int c = threadIdx.y + blockIdx.y*blockDim.y;
    int r = threadIdx.x + blockIdx.x*blockDim.x;

    if (r<rows && c<cols) {
        r += nhalos;
        c += nhalos;
        double mean = 0.;
        for (int i=r-nhalos; i<=r+nhalos; i++) {
             for (int j=c-nhalos; j<=c+nhalos; j++) {
                 double d  = 1.0*data[i*rowsize+j];
                 double e = exp(-(double)((i-r)*(i-r)+(j-c)*(j-c))/sig);
                 mean += d*e;
             }
        }
        smootheddata[r*rowsize + c] = (uint8_t)round(mean/(sig*pi));
    }
    return;
}

/* This time, use precomputed weights in global memory */
__global__ void cuda_smoothimage_weights
                                (uint8_t *data,
                                 uint8_t *smootheddata,
                                 double  *weights,
                                 const int nhalos, const int rows, const int cols) {

    const int rowsize=cols+2*nhalos;

    int r = threadIdx.y + blockIdx.y*blockDim.y;
    int c = threadIdx.x + blockIdx.x*blockDim.x;
    if (r<rows && c<cols) {
        r += nhalos;
        c += nhalos;
        double mean = 0.;
        for (int i=-nhalos; i<=+nhalos; i++) {
             for (int j=-nhalos; j<=+nhalos; j++) {
                 double d  = 1.0*data[(i+r)*rowsize+(j+c)];
                 mean += d*weights[(i+nhalos)*(2*nhalos+1)+(j+nhalos)];
             }
        }
        smootheddata[r*rowsize + c] = (uint8_t)round(mean);
    }
    return;
}

/* As above, but with worse memory access patterns*/
__global__ void cuda_smoothimage_weights_uncoallesced
                                (uint8_t *data,
                                 uint8_t *smootheddata,
                                 double  *weights,
                                 const int nhalos, const int rows, const int cols) {

    const int rowsize=cols+2*nhalos;

    int r = threadIdx.x + blockIdx.x*blockDim.x;
    int c = threadIdx.y + blockIdx.y*blockDim.y;

    if (r<rows && c<cols) {
        r += nhalos;
        c += nhalos;
        double mean = 0.;
        for (int i=-nhalos; i<=+nhalos; i++) {
             for (int j=-nhalos; j<=+nhalos; j++) {
                 double d  = 1.0*data[(i+r)*rowsize+(j+c)];
                 mean += d*weights[(i+nhalos)*(2*nhalos+1)+(j+nhalos)];
             }
        }
        smootheddata[r*rowsize + c] = (uint8_t)round(mean);
    }
    return;
}


/* Load precomputed weights into shared memory */
__global__ void cuda_smoothimage_weights_shared
                                (uint8_t *data,
                                 uint8_t *smootheddata,
                                 double  *gweights,
                                 const int nhalos, const int rows, const int cols) {

    extern __shared__ double weights[];
    const int rowsize=cols+2*nhalos;

    const int idx = threadIdx.x + threadIdx.y*blockDim.x;
    if (idx < (2*nhalos+1)*(2*nhalos+1))
        weights[idx] = gweights[idx];
    __syncthreads();

    // load global-memory weights into shared weights
    int r = threadIdx.y + blockIdx.y*blockDim.y;
    int c = threadIdx.x + blockIdx.x*blockDim.x;

    if (r<rows && c<cols) {
        r += nhalos;
        c += nhalos;
        double mean = 0.;
        for (int i=-nhalos; i<=+nhalos; i++) {
             for (int j=-nhalos; j<=+nhalos; j++) {
                 double d  = 1.0*data[(i+r)*rowsize+(j+c)];
                 mean += d*weights[(i+nhalos)*(2*nhalos+1)+(j+nhalos)];
             }
        }
        smootheddata[r*rowsize + c] = (uint8_t)round(mean);
    }
    return;
}

/* Calculate weights into shared memory */
__global__ void cuda_smoothimage_calc_weights_shared
                                (uint8_t *data,
                                 uint8_t *smootheddata,
                                 const int nhalos, const int rows, const int cols) {

    extern __shared__ double weights[];
    const int rowsize=cols+2*nhalos;

    if ((threadIdx.x < 2*nhalos+1) &&(threadIdx.y < 2*nhalos+1)) {
        const double sig = 2./9.*(nhalos*nhalos);
        const double pi = 4.*atan(1.);

        const int i=threadIdx.x - nhalos;
        const int j=threadIdx.y - nhalos;
        weights[threadIdx.y*(2*nhalos+1)+threadIdx.x] = exp(-(double)(i*i+j*j)/sig)/(sig*pi);
    }
    __syncthreads();

    // load global-memory weights into shared weights
    int r = threadIdx.y + blockIdx.y*blockDim.y;
    int c = threadIdx.x + blockIdx.x*blockDim.x;
    if (r<rows && c<cols) {
        r += nhalos;
        c += nhalos;
        double mean = 0.;
        for (int i=-nhalos; i<=+nhalos; i++) {
             for (int j=-nhalos; j<=+nhalos; j++) {
                 double d  = 1.0*data[(i+r)*rowsize+(j+c)];
                 mean += d*weights[(i+nhalos)*(2*nhalos+1)+(j+nhalos)];
             }
        }
        smootheddata[r*rowsize + c] = (uint8_t)round(mean);
    }
    return;
}


__global__ void cuda_smoothimage_tile_shared_smallblocks
                                (uint8_t *data,
                                 uint8_t *smootheddata,
                                 double  *weights,
                                 const int nhalos, const int rows, const int cols) {

    extern __shared__ uint8_t tile[];
    const int tilerowsize=blockDim.x;
    const int rowsize=cols+2*nhalos;

    int r = threadIdx.y + blockIdx.y*blockDim.y - 2*nhalos*blockIdx.y;
    int c = threadIdx.x + blockIdx.x*blockDim.x - 2*nhalos*blockIdx.x;

    tile[threadIdx.x + tilerowsize*threadIdx.y] = data[r*rowsize + c];
    __syncthreads();

    if (((r>nhalos) && r<(rows-nhalos)) && ((c>nhalos) & c<(cols-nhalos))) {
        if ((threadIdx.x > nhalos) && (threadIdx.x < blockDim.x - nhalos) &&
            (threadIdx.y > nhalos) && (threadIdx.y < blockDim.y - nhalos)) {
            double mean = 0.;
            for (int i=-nhalos; i<=+nhalos; i++) {
                 for (int j=-nhalos; j<=+nhalos; j++) {
                     double d  = 1.0*tile[(threadIdx.y+i)*tilerowsize+(threadIdx.x+j)];
                     mean += d*weights[(i+nhalos)*(2*nhalos+1)+(j+nhalos)];
                 }
            }
            smootheddata[r*rowsize + c] = (uint8_t)round(mean);
        }
    }

    return;
}


uint8_t charmean(uint8_t *data, const int r, const int c, const int nhalo, const int rowsize) {

    const double sig = 2./9.*(nhalo*nhalo);
    double pi = 4.*atan(1.);
    double mean=0.;
    for (int i=r-nhalo; i<=r+nhalo; i++) {
        for (int j=c-nhalo; j<=c+nhalo; j++) {
            double d  = 1.0*data[i*rowsize+j];
            double e = expf(-(double)((i-r)*(i-r)+(j-c)*(j-c))/sig);
            mean += d*e;
        }
    }

    mean /= sig*pi;

    return (uint8_t)roundf(mean);
}


void smoothimage(uint8_t *data, uint8_t *smootheddata, const int nhalos, const int rows, const int cols) {

    const int rowsize = rows+2*nhalos;

    for (int i=nhalos; i<nhalos+rows; i++) {
        for (int j=nhalos; j<nhalos+cols; j++) {
            smootheddata[i*rowsize+j] = charmean(data, i, j, nhalos, rowsize);
        }
    }
}


uint8_t charmean_weights(uint8_t *data, double *weights, const int r, const int c, const int nhalo, const int rowsize) {

    double mean=0.;
    for (int i=-nhalo; i<=+nhalo; i++) {
        for (int j=-nhalo; j<=+nhalo; j++) {
            double d  = 1.0*data[(i+r)*rowsize+(j+c)];
            double e = weights[(i+nhalo)*(2*nhalo+1)+(j+nhalo)];
            mean += d*e;
        }
    }

    return (uint8_t)roundf(mean);
}

void smoothimage_weights(uint8_t *data, uint8_t *smootheddata, double *weights, const int nhalos, const int rows, const int cols) {

    const int rowsize = rows+2*nhalos;

    for (int i=nhalos; i<nhalos+rows; i++) {
        for (int j=nhalos; j<nhalos+cols; j++) {
            smootheddata[i*rowsize+j] = charmean_weights(data, weights, i, j, nhalos, rowsize);
        }
    }
}


int pgmWriteWithHalo(char *fname, const int cols, const int rows,
                           const int max, const int nhalo, uint8_t **data);

int pgmReadWithHalo(char *fname, int *cols, int *rows,
                           int *max, int nhalo, uint8_t ***data);

uint8_t **char2darray(int n, int m);

void free2darray(uint8_t **array);

void setchar2darray(uint8_t **array, int n, int m, uint8_t val);

void tick(struct timeval *timer);

double tock(struct timeval *timer);

float calcdiff(const int nhalos, const int rows, const int cols,
        uint8_t** cudasmootheddata, uint8_t** smootheddata) {
    float abserr = 0.;
    for (int i = nhalos; i < rows + nhalos; i++) {
        for (int j = nhalos; j < cols + nhalos; j++) {
            abserr += fabs(cudasmootheddata[i][j] - smootheddata[i][j]);
        }
    }

    return abserr;
}

int main(int argc, char **argv) {
    
    if (argc != 6) {
        fprintf(stderr,"Usage: %s infile.pgm outfile-cpu.pgm outfile-gpu.pgm filtersize blocksize\n", argv[0]);
        fprintf(stderr,"       Smooths infile with median filter of size filtersize, writes to outfile.\n");
        exit(-1);
    }

    uint8_t **data;
    int  cols, rows, max;
    struct timeval cputimer, gputimer;

    int nhalos = atoi(argv[4]);
    const int defnhalos = 3;
    const int maxnhalos = 50;
    if (nhalos < 0 || nhalos > maxnhalos) {
        fprintf(stderr,"%s: filter size %s invalid; using default %d\n", argv[0], argv[3], defnhalos);
        nhalos = defnhalos;
    }

    int blocksize = atoi(argv[5]);
    const int defbsize = 8;
    const int maxbsize = 32;
    if (blocksize < 1 || blocksize > maxbsize) {
        fprintf(stderr,"%s: block size %s invalid; using default %d\n", argv[0], argv[4], defbsize);
        blocksize = defbsize;
    }

    int nread = pgmReadWithHalo(argv[1], &cols, &rows, &max, nhalos, &data);
    if (nread == 0) {
        fprintf(stderr,"%s: Error reading file %s\n", argv[0], argv[1]);
        exit(-2);
    }

    size_t datasize = (rows+2*nhalos)*(cols+2*nhalos)*sizeof(uint8_t);
    uint8_t **smootheddata     = char2darray(rows+2*nhalos, cols+2*nhalos);
    uint8_t **cudasmootheddata = char2darray(rows+2*nhalos, cols+2*nhalos);
    
    double *weightsd;
    double cputime, gputime;
    float  abserr;
    uint8_t *datad, *smoothedd;
    dim3 grid, block;

    tick(&cputimer);
    smoothimage(&(data[0][0]), &(smootheddata[0][0]), nhalos, rows, cols);
    cputime = tock(&cputimer);

    tick(&gputimer);
    CHK_CUDA( cudaMalloc(&datad, datasize ));
    CHK_CUDA( cudaMalloc(&smoothedd, datasize ));
    CHK_CUDA(cudaMemcpy(datad, &(data[0][0]), datasize, cudaMemcpyHostToDevice));

    grid = make_uint3((rows + blocksize - 1) / blocksize, (cols + blocksize - 1) / blocksize, 1);
    block = make_uint3(blocksize, blocksize, 1);
    cuda_smoothimage<<<grid, block>>>(datad, smoothedd, nhalos, rows, cols);
    CHK_ERROR();

    CHK_CUDA(cudaMemcpy(&(cudasmootheddata[0][0]), smoothedd, datasize, cudaMemcpyDeviceToHost));
    CHK_CUDA( cudaFree(datad));
    CHK_CUDA( cudaFree(smoothedd));
    gputime = tock(&gputimer);
    abserr = calcdiff(nhalos, rows, cols, cudasmootheddata, smootheddata);


    tick(&gputimer);
    CHK_CUDA( cudaMalloc(&datad, datasize ));
    CHK_CUDA( cudaMalloc(&smoothedd, datasize ));
    CHK_CUDA(cudaMemcpy(datad, &(data[0][0]), datasize, cudaMemcpyHostToDevice));

    grid = make_uint3((rows + blocksize - 1) / blocksize, (cols + blocksize - 1) / blocksize, 1);
    block = make_uint3(blocksize, blocksize, 1);
    cuda_smoothimage<<<grid, block>>>(datad, smoothedd, nhalos, rows, cols);
    CHK_ERROR();

    CHK_CUDA(cudaMemcpy(&(cudasmootheddata[0][0]), smoothedd, datasize, cudaMemcpyDeviceToHost));
    CHK_CUDA( cudaFree(datad));
    CHK_CUDA( cudaFree(smoothedd));
    gputime = tock(&gputimer);
    abserr = calcdiff(nhalos, rows, cols, cudasmootheddata, smootheddata);

    printf("Image size = %d x %d, Number of blocks = %d x %d.\n", rows, cols, grid.x, grid.y);
    printf("CPU time = %lg millisec.\n", cputime*1000.);
    printf("GPU time = %lg millisec.\n", gputime*1000.);
    printf("CUDA and CPU results differ by %lf\n", abserr);


    tick(&gputimer);
    CHK_CUDA( cudaMalloc(&datad, datasize ));
    CHK_CUDA( cudaMalloc(&smoothedd, datasize ));
    CHK_CUDA(cudaMemcpy(datad, &(data[0][0]), datasize, cudaMemcpyHostToDevice));

    grid = make_uint3((cols + blocksize - 1) / blocksize, (rows + blocksize - 1) / blocksize, 1);
    block = make_uint3(blocksize, blocksize, 1);
    cuda_smoothimage_uncoallesced<<<grid, block>>>(datad, smoothedd, nhalos, rows, cols);
    CHK_ERROR();

    CHK_CUDA(cudaMemcpy(&(cudasmootheddata[0][0]), smoothedd, datasize, cudaMemcpyDeviceToHost));
    CHK_CUDA( cudaFree(datad));
    CHK_CUDA( cudaFree(smoothedd));
    gputime = tock(&gputimer);
    abserr = calcdiff(nhalos, rows, cols, cudasmootheddata, smootheddata);

    printf("--- uncoallesced ---\n");
    printf("GPU time = %lg millisec.\n", gputime*1000.);
    printf("CUDA and CPU results differ by %lf\n", abserr);


    /* precompute weights */
    const int weightsize = (2 * nhalos + 1) * (2 * nhalos + 1) * sizeof(double);
    double* weights = (double*) (malloc(weightsize));
    const double pi = 4. * atan(1.);
    const double sig = 2. / 9. * (nhalos * nhalos);
    for (int i = -nhalos; i <= nhalos; i++)
        for (int j = -nhalos; j <= nhalos; j++)
            weights[(i + nhalos) * (2 * nhalos + 1) + (j + nhalos)] = exp(
                    -(double) ((i * i + j * j)) / sig) / (sig * pi);

    /* do timings with precomputed weights */
    tick(&cputimer);
    smoothimage_weights(&(data[0][0]), &(smootheddata[0][0]), weights, nhalos,
            rows, cols);
    cputime = tock(&cputimer);
    tick(&gputimer);
    CHK_CUDA( cudaMalloc(&datad, datasize ));
    CHK_CUDA( cudaMalloc(&smoothedd, datasize ));
    CHK_CUDA( cudaMalloc(&weightsd, weightsize));
    CHK_CUDA(cudaMemcpy(datad, &(data[0][0]), datasize, cudaMemcpyHostToDevice));
    CHK_CUDA(cudaMemcpy(weightsd, weights, weightsize, cudaMemcpyHostToDevice));

    grid = make_uint3((rows + blocksize - 1) / blocksize,
            (cols + blocksize - 1) / blocksize, 1);
    block = make_uint3(blocksize, blocksize, 1);
    cuda_smoothimage_weights<<<grid, block>>>(datad, smoothedd, weightsd, nhalos, rows, cols);
    CHK_ERROR();

    CHK_CUDA(cudaMemcpy(&(cudasmootheddata[0][0]), smoothedd, datasize, cudaMemcpyDeviceToHost));
    CHK_CUDA( cudaFree(datad));
    CHK_CUDA( cudaFree(smoothedd));
    gputime = tock(&gputimer);
    abserr = calcdiff(nhalos, rows, cols, cudasmootheddata, smootheddata);


    tick(&gputimer);
    CHK_CUDA( cudaMalloc(&datad, datasize ));
    CHK_CUDA( cudaMalloc(&smoothedd, datasize ));
    CHK_CUDA( cudaMalloc(&weightsd, weightsize));
    CHK_CUDA( cudaMemcpy(datad, &(data[0][0]), datasize, cudaMemcpyHostToDevice));
    CHK_CUDA( cudaMemcpy(weightsd, weights, weightsize, cudaMemcpyHostToDevice));
    grid = make_uint3((rows + blocksize - 1) / blocksize,
            (cols + blocksize - 1) / blocksize, 1);
    block = make_uint3(blocksize, blocksize, 1);
    cuda_smoothimage_weights<<<grid, block>>>(datad, smoothedd, weightsd, nhalos, rows, cols);
    CHK_ERROR();

    CHK_CUDA( cudaMemcpy(&(cudasmootheddata[0][0]), smoothedd, datasize, cudaMemcpyDeviceToHost));
    CHK_CUDA( cudaFree(datad));
    CHK_CUDA( cudaFree(smoothedd));
    gputime = tock(&gputimer);

    abserr = calcdiff(nhalos, rows, cols, cudasmootheddata, smootheddata);

    printf("--- precomputing weights ---\n");
    printf("Image size = %d x %d, Number of blocks = %d x %d.\n", rows, cols, grid.x, grid.y);
    printf("CPU time = %lg millisec.\n", cputime*1000.);
    printf("GPU time = %lg millisec.\n", gputime*1000.);
    printf("CUDA and CPU results differ by %lf\n", abserr);

    int nwrote = pgmWriteWithHalo(argv[2], cols, rows, max, nhalos, smootheddata);

    if (nwrote == 0) {
        fprintf(stderr,"%s: Error writing file %s\n", argv[0], argv[2]);
        free2darray(smootheddata);
        free2darray(data);
        exit(-3);
    }


    tick(&gputimer);
    CHK_CUDA( cudaMalloc(&datad,     datasize ) );
    CHK_CUDA( cudaMalloc(&smoothedd, datasize ) );
    CHK_CUDA( cudaMalloc(&weightsd,  weightsize));

    CHK_CUDA( cudaMemcpy(datad, &(data[0][0]), datasize, cudaMemcpyHostToDevice) );
    CHK_CUDA( cudaMemcpy(weightsd, weights, weightsize,  cudaMemcpyHostToDevice) );
    grid   = make_uint3( (cols+blocksize-1)/blocksize, (rows+blocksize-1)/blocksize,  1);
    block  = make_uint3( blocksize, blocksize, 1);

    cuda_smoothimage_weights_uncoallesced<<<grid, block>>>(datad, smoothedd, weightsd, nhalos, rows, cols);
    CHK_ERROR();

    CHK_CUDA( cudaMemcpy(&(cudasmootheddata[0][0]), smoothedd, datasize, cudaMemcpyDeviceToHost) );
    CHK_CUDA( cudaFree(datad) );
    CHK_CUDA( cudaFree(smoothedd) );
    gputime = tock(&gputimer);

    abserr = calcdiff(nhalos, rows, cols, cudasmootheddata, smootheddata);

    printf("--- uncoallesced ---\n");
    printf("GPU time = %lg millisec.\n", gputime*1000.);
    printf("CUDA and CPU results differ by %lf\n", abserr);


    printf("--- using shared memory ---\n");
    /* now with shared memory weights */
    tick(&gputimer);
    CHK_CUDA( cudaMalloc(&datad,     datasize ) );
    CHK_CUDA( cudaMalloc(&smoothedd, datasize ) );
    CHK_CUDA( cudaMalloc(&weightsd,  weightsize));

    CHK_CUDA( cudaMemcpy(datad, &(data[0][0]), datasize, cudaMemcpyHostToDevice) );
    CHK_CUDA( cudaMemcpy(weightsd, weights, weightsize,  cudaMemcpyHostToDevice) );
    grid   = make_uint3( (cols+blocksize-1)/blocksize, (rows+blocksize-1)/blocksize,  1);
    block  = make_uint3( blocksize, blocksize, 1);

    cuda_smoothimage_weights_shared<<<grid, block, weightsize>>>(datad, smoothedd, weightsd, nhalos, rows, cols);
    CHK_ERROR();

    CHK_CUDA( cudaMemcpy(&(cudasmootheddata[0][0]), smoothedd, datasize, cudaMemcpyDeviceToHost) );
    CHK_CUDA( cudaFree(datad) );
    CHK_CUDA( cudaFree(smoothedd) );
    CHK_CUDA( cudaFree(weightsd) );
    gputime = tock(&gputimer);

    abserr = calcdiff(nhalos, rows, cols, cudasmootheddata, smootheddata);

    printf("GPU time = %lg millisec (shared weights).\n", gputime*1000.);
    printf("Shared weights and CPU results differ by %lf\n", abserr);


    /* now with shared memory calculated weights */
    tick(&gputimer);
    CHK_CUDA( cudaMalloc(&datad,     datasize ) );
    CHK_CUDA( cudaMalloc(&smoothedd, datasize ) );

    CHK_CUDA( cudaMemcpy(datad, &(data[0][0]), datasize, cudaMemcpyHostToDevice) );
    grid   = make_uint3( (rows+blocksize-1)/blocksize, (cols+blocksize-1)/blocksize,  1);
    block  = make_uint3( blocksize, blocksize, 1);

    cuda_smoothimage_calc_weights_shared<<<grid, block, weightsize>>>(datad, smoothedd, nhalos, rows, cols);
    CHK_ERROR();

    CHK_CUDA( cudaMemcpy(&(cudasmootheddata[0][0]), smoothedd, datasize, cudaMemcpyDeviceToHost) );
    CHK_CUDA( cudaFree(datad) );
    CHK_CUDA( cudaFree(smoothedd) );
    gputime = tock(&gputimer);

    abserr = calcdiff(nhalos, rows, cols, cudasmootheddata, smootheddata);

    printf("GPU time = %lg millisec (calced shared weights).\n", gputime*1000.);
    printf("Shared weights and CPU results differ by %lf\n", abserr);

    /* now with shared memory tile */

    tick(&gputimer);
    CHK_CUDA( cudaMalloc(&datad,     datasize ) );
    CHK_CUDA( cudaMalloc(&smoothedd, datasize ) );
    CHK_CUDA( cudaMalloc(&weightsd,  weightsize));

    CHK_CUDA( cudaMemcpy(datad, &(data[0][0]), datasize, cudaMemcpyHostToDevice) );
    CHK_CUDA( cudaMemcpy(weightsd, weights, weightsize,  cudaMemcpyHostToDevice) );

    const int innerblocksize=(blocksize-2*nhalos);
    grid   = make_uint3( (rows+innerblocksize-1)/innerblocksize, (cols+innerblocksize-1)/innerblocksize,  1);
    block  = make_uint3( blocksize, blocksize, 1);
    const int blockdatasize=blocksize*blocksize*sizeof(uint8_t);

    cuda_smoothimage_tile_shared_smallblocks<<<grid, block, blockdatasize>>>(datad, smoothedd, weightsd, nhalos, rows, cols);
    CHK_ERROR();

    CHK_CUDA( cudaMemcpy(&(cudasmootheddata[0][0]), smoothedd, datasize, cudaMemcpyDeviceToHost) );
    CHK_CUDA( cudaFree(datad) );
    CHK_CUDA( cudaFree(smoothedd) );
    gputime = tock(&gputimer);

    abserr = calcdiff(nhalos, rows, cols, cudasmootheddata, smootheddata);

    printf("GPU time = %lg millisec (shared image tiles).\n", gputime*1000.);
    printf("Shared weights and CPU results differ by %lf\n", abserr);

    nwrote = pgmWriteWithHalo(argv[3], cols, rows, max, nhalos, cudasmootheddata);

    if (nwrote == 0) {
        fprintf(stderr,"%s: Error writing file %s\n", argv[0], argv[3]);
        free2darray(cudasmootheddata);
        free2darray(data);
        exit(-3);
    }

    free2darray(cudasmootheddata);
    free2darray(smootheddata); 
    free2darray(data); 
    return 0;
}


/*
 * Utility routines used by the main program for allocating 2d uint8_tacter
 * arrays, and reading/writing image files in PGM format.
 */

#define PGMMAGIC "P5"

void getNonCommentLine(FILE *in, char *line, const int len) {
    do {
        fgets(line, len, in);
    } while (line[0] == '#');
}

void scanHeader(FILE *in, char *magic, int *col, int *row, int *max) {
    const int len=128;
    char dummy[len+1];

    getNonCommentLine(in, dummy, len);
    sscanf(dummy, "%s\n", magic);

    getNonCommentLine(in, dummy, len);
    sscanf(dummy, "%d %d\n", col, row);
    
    getNonCommentLine(in, dummy, len);
    sscanf(dummy, "%d\n", max);
}

int pgmReadHeader(char *fname, char *magic, int *col, int *row, int *max) {
    
    FILE *in = fopen(fname,"r");
    if (!in) 
        return -1;

    scanHeader(in, magic, col, row, max);
    fclose(in);

    return 0;
}

void skipheader(FILE *in) {
    char magic[3];
    int col, row, max; 
    scanHeader(in, magic, &col, &row, &max);
}

#include <string.h>

int correctmagic(char *magic) {
    return !strncmp(magic, PGMMAGIC, 2);
}

int pgmReadData (char *fname, const int cols, const int rows, uint8_t *data) {
    size_t nread;
    
    FILE *in = fopen(fname,"r");
    if (!in) 
        return -1;

    skipheader(in);

    nread = fread(data, sizeof(uint8_t), cols*rows, in);

    fclose(in);
    return nread;
}


int pgmReadDataWithHalo(char *fname, const int cols, const int rows,
                           int nhalo, uint8_t **data) {
    size_t nread = 0;
    
    FILE *in = fopen(fname,"r");
    if (!in) 
        return -1;

    skipheader(in);

    for (int row=0; row<rows; row++) 
        nread += fread(&(data[row+nhalo][nhalo]), sizeof(uint8_t), cols, in);

    fclose(in);
    return nread;
}

void writeHeader(FILE *out, int cols, int rows, int max) {

    fprintf(out, "%s\n", PGMMAGIC);
    fprintf(out, "%d %d\n", cols, rows);
    fprintf(out, "%d\n", max);

}

int pgmWrite(char *fname, const int cols, const int rows,
                     const int max, uint8_t *data) {
    size_t nwrote;
    
    FILE *out = fopen(fname,"wb");
    if (!out) 
        return -1;

    writeHeader(out, cols, rows, max);
    nwrote = fwrite(data, sizeof(uint8_t), cols*rows, out);

    fclose(out);
    return nwrote;
}


int pgmWriteWithHalo(char *fname, const int cols, const int rows,
                           const int max, const int nhalo, uint8_t **data) {
    FILE *out = fopen(fname,"wb");
    
    if (!out) 
        return -1;

    writeHeader(out, cols, rows, max);

    size_t nwrote = 0;
    for (int row=0; row<rows; row++) 
        nwrote += fwrite(&(data[row+nhalo][nhalo]), sizeof(uint8_t), cols, out);

    fclose(out);
    return nwrote;
}

size_t pgmRead(char *fname, int *cols, int *rows, int *max, uint8_t ***data) {

    char magic[6];
    int res = pgmReadHeader(fname, magic, cols, rows, max);
    if (res < 0)
        return 0;

    if (!correctmagic(magic))
        return 0;

    *data = char2darray(*rows, *cols);
    return pgmReadData (fname, *cols, *rows, &((*data)[0][0]));
}


int pgmReadWithHalo(char *fname, int *cols, int *rows,
                           int *max, int nhalo, uint8_t ***data) {

    char magic[6];
    int res = pgmReadHeader(fname, magic, cols, rows, max);
    if (res < 0)
        return 0;

    if (!correctmagic(magic))
        return 0;

    *data = char2darray((*rows)+2*nhalo, (*cols)+2*nhalo);
    setchar2darray(*data, (*rows)+2*nhalo, (*cols)+2*nhalo, 0);

    return pgmReadDataWithHalo (fname, *cols, *rows, nhalo, *data);
}

uint8_t **char2darray(int n, int m) {
    uint8_t *data = (uint8_t *)malloc(n*m*sizeof(uint8_t));
    uint8_t **array = (uint8_t **)malloc(n*sizeof(uint8_t*));
    for (int i=0; i<n; i++)
        array[i] = &(data[i*m]);

    return array;
}

void free2darray(uint8_t **array) {
    free(array[0]);
    free(array);
}

void setchar2darray(uint8_t **array, int n, int m, uint8_t val) {
    memset(array[0], val, n*m);
}


void tick(struct timeval *timer) {
    gettimeofday(timer, NULL);
}

double tock(struct timeval *timer) {
    struct timeval now;
    gettimeofday(&now, NULL);
    return (now.tv_usec-timer->tv_usec)/1.0e6 + (now.tv_sec - timer->tv_sec);
}
